from __future__ import annotations

import io
import gc
import sys
import yaml
import logging
import vapoursynth as vs
from pathlib import Path
from itertools import count
from os.path import expandvars, expanduser
from traceback import FrameSummary, TracebackException
from typing import Any, cast, List, Mapping, Tuple, Dict

from PyQt5.QtCore import pyqtSignal, QRectF, QEvent
from PyQt5.QtGui import QCloseEvent, QColorSpace, QPalette, QShowEvent, QPixmap, QMoveEvent
from PyQt5.QtWidgets import QLabel, QApplication, QGraphicsScene, QOpenGLWidget, QSizePolicy, QGraphicsView

from ..toolbars import Toolbars
from ..models import VideoOutputs
from ..core.vsenv import get_policy
from ..utils import fire_and_forget, set_status_label
from ..core.custom import StatusBar, GraphicsView, GraphicsImageItem, DragNavigator
from ..core import AbstractMainWindow, Frame, VideoOutput, Time, try_load, VBoxLayout, ExtendedWidget, ViewMode
from ..core.types.enums import Matrix, Transfer, Primaries, ColorRange, ChromaLocation

from .timeline import Timeline
from .settings import MainSettings, WindowSettings
from .dialog import ScriptErrorDialog, SettingsDialog

if sys.platform == 'win32':
    import win32gui

    try:
        from PIL import _imagingcms
    except ImportError:
        _imagingcms = None

try:
    from yaml import CLoader as yaml_Loader, CDumper as yaml_Dumper
except ImportError:
    from yaml import Loader as yaml_Loader, Dumper as yaml_Dumper


class MainWindow(AbstractMainWindow):
    VSP_DIR_NAME = '.vspreview'
    VSP_GLOBAL_DIR_NAME = Path(
        expandvars('%APPDATA%') if sys.platform == "win32" else expanduser('~/.config')
    )
    # used for formats with subsampling
    VS_OUTPUT_MATRIX = Matrix.BT709
    VS_OUTPUT_TRANSFER = Transfer.BT709
    VS_OUTPUT_PRIMARIES = Primaries.BT709
    VS_OUTPUT_RANGE = ColorRange.LIMITED
    VS_OUTPUT_CHROMALOC = ChromaLocation.LEFT
    VS_OUTPUT_RESIZER_KWARGS = {
        'dither_type': 'error_diffusion',
    }
    VSP_VERSION = 2.1
    BREAKING_CHANGES_VERSIONS: List[float] = []

    # status bar
    def STATUS_FRAME_PROP(self, prop: Any) -> str:
        return 'Type: %s' % (prop['_PictType'].decode('utf-8') if '_PictType' in prop else '?')

    EVENT_POLICY = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    storable_attrs = ('settings', 'toolbars')

    __slots__ = (
        *storable_attrs, 'app', 'display_scale', 'clipboard',
        'script_path', 'timeline', 'main_layout',
        'graphics_scene', 'graphics_view', 'script_error_dialog',
        'central_widget', 'statusbar', 'storage_not_found',
        'current_storage_path', 'opengl_widget', 'drag_navigator'
    )

    # emit when about to reload a script: clear all existing references to existing clips.
    reload_signal = pyqtSignal()
    reload_before_signal = pyqtSignal()
    reload_after_signal = pyqtSignal()
    toolbars: Toolbars

    def __init__(self, config_dir: Path) -> None:
        super().__init__()

        self.settings = MainSettings()

        # logging
        logging.basicConfig(format='{asctime}: {levelname}: {message}', style='{', level=self.settings.LOG_LEVEL)
        logging.Formatter.default_msec_format = '%s.%03d'

        self.current_config_dir = config_dir / self.VSP_DIR_NAME
        self.global_config_dir = self.VSP_GLOBAL_DIR_NAME / self.VSP_DIR_NAME
        self.global_storage_path = self.global_config_dir / '.global.yml'

        self.app = QApplication.instance()
        assert self.app

        if self.settings.dark_theme_enabled:
            try:
                from qdarkstyle import load_stylesheet_pyqt5
            except ImportError:
                self.self.settings.dark_theme_enabled = False
            else:
                self.app.setStyleSheet(self.patch_dark_stylesheet(load_stylesheet_pyqt5()))
                self.ensurePolished()

        self.display_scale = self.app.primaryScreen().logicalDotsPerInch() / self.settings.base_ppi
        self.setWindowTitle('VSPreview')

        desktop_size = self.app.primaryScreen().size()

        self.move(int(desktop_size.width() * 0.15), int(desktop_size.height() * 0.075))
        self.setup_ui()
        self.storage_not_found = False
        self.script_globals: Dict[str, Any] = dict()

        # global
        self.clipboard = self.app.clipboard()
        self.external_args: List[Tuple[str, str]] = []
        self.script_path = Path()
        self.script_exec_failed = False
        self.current_storage_path = Path()

        # graphics view
        self.graphics_scene = QGraphicsScene(self)
        self.graphics_view.setScene(self.graphics_scene)
        self.opengl_widget = None

        if self.settings.opengl_rendering_enabled:
            self.opengl_widget = QOpenGLWidget()
            self.graphics_view.setViewport(self.opengl_widget)

        self.graphics_view.wheelScrolled.connect(self.on_wheel_scrolled)

        self.graphics_view.registerReloadEvents(self)

        # timeline
        self.timeline.clicked.connect(self.on_timeline_clicked)

        # display profile
        self.display_profile: QColorSpace | None = None
        self.current_screen = 0

        # init toolbars and outputs
        self.app_settings = SettingsDialog(self)
        self.toolbars = Toolbars(self)

        for toolbar in self.toolbars:
            self.main_layout.addWidget(toolbar)
            self.toolbars.main.layout().addWidget(toolbar.toggle_button)

        self.app_settings.tab_widget.setUsesScrollButtons(False)
        self.app_settings.setMinimumWidth(
            int(len(self.toolbars) * 1.05 * self.app_settings.tab_widget.geometry().width() / 2)
        )

        self.current_viewmode = ViewMode.NORMAL

        self.set_qobject_names()
        self.setObjectName('MainWindow')

    def setup_ui(self) -> None:
        self.central_widget = ExtendedWidget(self)
        self.main_layout = VBoxLayout(self.central_widget)
        self.setCentralWidget(self.central_widget)

        self.graphics_view = GraphicsView(self.central_widget)
        self.graphics_view.setBackgroundBrush(self.palette().brush(QPalette.Window))
        self.graphics_view.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.graphics_view.setDragMode(QGraphicsView.ScrollHandDrag)

        self.drag_navigator = DragNavigator(self, self.graphics_view)

        self.timeline = Timeline(self.central_widget)

        self.main_layout.addWidgets([self.graphics_view, self.timeline])

        # status bar
        self.statusbar = StatusBar(self.central_widget)

        for name in self.statusbar.label_names:
            self.statusbar.__setattr__(name, QLabel(self.central_widget))

        self.statusbar.addWidgets([
            getattr(self.statusbar, name) for name in self.statusbar.label_names
        ])

        self.statusbar.addPermanentWidget(self.statusbar.label)

        self.setStatusBar(self.statusbar)

        # dialogs
        self.script_error_dialog = ScriptErrorDialog(self)

    def patch_dark_stylesheet(self, stylesheet: str) -> str:
        return stylesheet + 'QGraphicsView { border: 0px; padding: 0px; }'

    def load_script(
        self, script_path: Path, external_args: List[Tuple[str, str]] | None = None, reloading: bool = False,
        start_frame: int | None = None
    ) -> None:
        self.external_args = external_args or []

        self.toolbars.playback.stop()
        self.setWindowTitle('VSPreview: %s %s' % (script_path, self.external_args))

        self.statusbar.label.setText('Evaluating')
        self.script_path = script_path

        sys.path.append(str(self.script_path.parent))

        # Rewrite args so external args will be forwarded correctly
        try:
            argv_orig = sys.argv
            sys.argv = [script_path.name]
        except AttributeError:
            pass

        self.script_globals.clear()
        self.script_globals = dict([('__file__', sys.argv[0])] + self.external_args)

        try:
            ast_compiled = compile(self.script_path.read_bytes(), sys.argv[0], 'exec', optimize=2)

            exec(ast_compiled, self.script_globals)
        except BaseException as e:
            logging.error(e)

            te = TracebackException.from_exception(e)
            # remove the first stack frame, which contains our exec() invocation
            del te.stack[0]

            # replace <string> with script path only for the first stack frames
            # in order to keep intact exec() invocations down the stack
            # that we're not concerned with
            for i, frame in enumerate(te.stack):
                if frame.filename == '<string>':
                    te.stack[i] = FrameSummary(
                        str(self.script_path), frame.lineno, frame.name
                    )
                else:
                    break
            logging.error(''.join(te.format()))

            self.script_exec_failed = True
            return self.handle_script_error(
                '\n'.join([
                    'An error occured while evaluating script:',
                    str(e), 'See console output for details.'
                ])
            )
        finally:
            sys.argv = argv_orig
            sys.path.pop()

        if len(vs.get_outputs()) == 0:
            logging.error('Script has no outputs set.')
            self.script_exec_failed = True
            self.handle_script_error('Script has no outputs set.')
            return

        self.script_exec_failed = False
        self.current_storage_path = (self.current_config_dir / self.script_path.stem).with_suffix('.yml')

        self.storage_not_found = not self.current_storage_path.exists()

        if self.storage_not_found:
            self.load_storage()

        if not reloading:
            self.toolbars.main.rescan_outputs()
            self.toolbars.playback.rescan_outputs()

        if not self.storage_not_found:
            self.load_storage()

        self.change_video_viewmode(self.current_viewmode)

        self.toolbars.misc.autosave_timer.start(round(float(self.settings.autosave_interval) * 1000))

        if not reloading:
            self.switch_output(self.settings.output_index)
            if start_frame is not None:
                self.switch_frame(Frame(start_frame))

    @set_status_label('Loading...')
    def load_storage(self) -> None:
        storage_paths = [self.global_storage_path, self.current_storage_path]

        if self.storage_not_found:
            logging.info('No storage found. Using defaults.')

            if not self.global_storage_path.exists():
                return

            storage_paths = storage_paths[:1]

        storage_contents = ''
        broken_storage = False
        global_length = 0
        for i, storage_path in enumerate(storage_paths):
            try:
                with io.open(storage_path, 'r', encoding='utf-8') as storage_file:
                    version = storage_file.readline()
                    if 'Version' not in version or any({
                        version.endswith(f'@{v}') for v in self.BREAKING_CHANGES_VERSIONS
                    }):
                        raise FileNotFoundError

                    storage_contents += storage_file.read()
                    storage_contents += '\n'

                    if i == 0:
                        global_length = storage_contents.count('\n')
            except FileNotFoundError:
                if self.settings.force_old_storages_removal:
                    storage_path.unlink()
                    broken_storage = True
                else:
                    logging.warning(
                        '\n\tThe storage was created on an old version of VSPreview.'
                        '\n\tSave any scening or other important info and delete it.'
                        '\n\tIf you want the program to silently delete old storages, go into settings.'
                    )
                    sys.exit(1)

        if broken_storage:
            return

        loader = yaml_Loader(storage_contents)
        try:
            loader.get_single_data()
        except yaml.YAMLError as exc:
            if isinstance(exc, yaml.MarkedYAMLError):
                if exc.problem_mark:
                    line = exc.problem_mark.line + 1
                    isglobal = line <= global_length
                    if not isglobal:
                        line -= global_length
                    logging.warning(
                        'Storage ({}) parsing failed on line {} column {}. \n({})\n Exiting...'
                        .format(
                            'Global' if isglobal else 'Local',
                            line, exc.problem_mark.column + 1,
                            str(exc).partition('in "<unicode string>"')[0].strip()
                        )
                    )
                    sys.exit(1)
            else:
                logging.warning('Storage parsing failed. Using defaults.')
        finally:
            loader.dispose()

        if self.settings.color_management:
            assert self.app
            self.current_screen = self.app.desktop().screenNumber(self)
            self.update_display_profile()

    @fire_and_forget
    @set_status_label('Saving...')
    def dump_storage_async(self) -> None:
        self.dump_storage()

    def dump_storage(self, manually: bool = False) -> None:
        if self.script_exec_failed:
            return

        self.current_config_dir.mkdir(0o777, True, True)
        self.global_config_dir.mkdir(0o777, True, True)

        backup_paths = [
            self.current_storage_path.with_suffix(f'.old{i}.yml')
            for i in range(self.toolbars.misc.settings.STORAGE_BACKUPS_COUNT, 0, -1)
        ] + [self.current_storage_path]

        for src_path, dest_path in zip(backup_paths[1:], backup_paths[:-1]):
            if src_path.exists():
                src_path.replace(dest_path)

        storage_dump = self._dump_serialize(self._serialize_data()).splitlines()

        idx = next(idx for (line, idx) in zip(storage_dump[2:], count(2)) if not line.startswith(' '))

        version = f'# Version@{self.VSP_VERSION}'

        with io.open(self.global_storage_path, 'w', encoding='utf-8') as global_file:
            global_file.writelines(
                '\n'.join([version, '# Global VSPreview storage for settings'] + storage_dump[:idx])
            )

        with io.open(self.current_storage_path, 'w', encoding='utf-8') as current_file:
            current_file.writelines(
                '\n'.join([
                    version,
                    f'# VSPreview script storage for: {self.script_path}',
                    f'# Global setting saved at path: {self.global_storage_path}'
                ] + storage_dump[idx:])
            )

        if manually:
            self.show_message('Saved successfully')

    def _serialize_data(self) -> Any:
        # idk how to explain how this work,
        # but i'm referencing settings objects before in the dict
        # so the yaml serializer will reference the same objects after (in toolbars),
        # which really are the original objects, to those copied in _globals :poppo:
        data = cast(dict, self.__getstate__())
        data['_globals'] = {
            'settings': data['settings'],
            'window_settings': data['window_settings']
        }

        data['_globals']['toolbars'] = data['toolbars'].__getstate__()
        gtoolbars = data['_globals']['toolbars']

        for toolbar_name in gtoolbars:
            gtoolbars[toolbar_name].clear()
            gtoolbars[toolbar_name] = getattr(data['toolbars'], toolbar_name).settings

        data['_toolbars_settings'] = [None] * (len(gtoolbars) + 1)
        for i, toolbar_name in enumerate(gtoolbars.keys()):
            data['_toolbars_settings'][i] = gtoolbars[toolbar_name]
        data['_toolbars_settings'][-1] = data['_globals']['window_settings']

        return data

    def _dump_serialize(self, data: Any) -> str:
        storage_dump = io.StringIO()

        dumper = yaml_Dumper(
            storage_dump, default_style=None, default_flow_style=False,
            canonical=None, indent=4, width=120, allow_unicode=True,
            line_break='\n', encoding='utf-8', version=None, tags=None,
            explicit_start=None, explicit_end=None, sort_keys=True
        )
        try:
            dumper.open()
            dumper.represent(data)
            dumper.close()
        finally:
            dumper.dispose()

        return storage_dump.getvalue()

    def init_outputs(self) -> None:
        self.graphics_scene.clear()

        for output in self.outputs:
            raw_frame_item = self.graphics_scene.addPixmap(QPixmap())
            raw_frame_item.hide()

            output.graphics_scene_item = GraphicsImageItem(raw_frame_item)

    def reload_script(self) -> None:
        self.reload_before_signal.emit()

        self.dump_storage()

        vs.clear_outputs()
        self.graphics_scene.clear()

        self.outputs.clear()
        get_policy().reload_core()
        gc.collect(generation=0)
        gc.collect(generation=1)
        gc.collect(generation=2)

        self.load_script(self.script_path, reloading=True)

        self.reload_after_signal.emit()

        self.show_message('Reloaded successfully')

    def switch_frame(
        self, pos: Frame | Time | int | None, *, render_frame: bool | Tuple[vs.VideoFrame, vs.VideoFrame | None] = True
    ) -> None:
        if pos is None:
            logging.debug('switch_frame: position is None!')
            return

        frame = Frame(min(max(0, int(Frame(pos))), int(self.current_output.total_frames)))

        if self.current_output.last_showed_frame == frame > self.current_output.end_frame:
            return

        if render_frame:
            if isinstance(render_frame, bool):
                self.current_output.render_frame(frame, output_colorspace=self.display_profile)
            else:
                self.current_output.render_frame(frame, *render_frame, output_colorspace=self.display_profile)

        self.current_output.last_showed_frame = frame

        self.timeline.set_position(frame)

        for toolbar in self.toolbars:
            toolbar.on_current_frame_changed(frame)

        self.statusbar.frame_props_label.setText(self.STATUS_FRAME_PROP(self.current_output.props))

    def switch_output(self, value: int | VideoOutput) -> None:
        if len(self.outputs) == 0:
            return

        if isinstance(value, VideoOutput):
            index = self.outputs.index_of(value)
        else:
            index = value

        if index < 0:
            index = len(self.outputs) + index

        if index < 0 or index >= len(self.outputs):
            return

        prev_index = self.toolbars.main.outputs_combobox.currentIndex()

        self.toolbars.playback.stop()

        # current_output relies on outputs_combobox
        self.toolbars.main.on_current_output_changed(index, prev_index)
        self.timeline.set_end_frame(self.current_output.end_frame)

        if self.current_output.last_showed_frame:
            self.switch_frame(self.current_output.last_showed_frame)
        else:
            self.switch_frame(0)

        for output in self.outputs:
            output.graphics_scene_item.hide()

        self.current_output.graphics_scene_item.show()
        self.graphics_scene.setSceneRect(QRectF(self.current_output.graphics_scene_item.pixmap().rect()))
        self.timeline.update_notches()

        for toolbar in self.toolbars[1:]:
            toolbar.on_current_output_changed(index, prev_index)

        self.update_statusbar_output_info()

    @property
    def current_output(self) -> VideoOutput:
        return cast(VideoOutput, self.toolbars.main.outputs_combobox.currentData())

    @current_output.setter
    def current_output(self, value: VideoOutput) -> None:
        self.switch_output(self.outputs.index_of(value))

    @property
    def outputs(self) -> VideoOutputs:
        return self.toolbars.main.outputs

    def handle_script_error(self, message: str) -> None:
        self.script_error_dialog.label.setText(message)
        self.script_error_dialog.open()

    def on_wheel_scrolled(self, steps: int) -> None:
        new_index = self.toolbars.main.zoom_combobox.currentIndex() + steps
        if new_index < 0:
            new_index = 0
        elif new_index >= len(self.settings.zoom_levels):
            new_index = len(self.settings.zoom_levels) - 1
        self.toolbars.main.zoom_combobox.setCurrentIndex(new_index)

    def on_timeline_clicked(self, start: int) -> None:
        if self.toolbars.playback.play_timer.isActive():
            self.toolbars.playback.stop()
            self.switch_frame(start)
            self.toolbars.playback.play()
        else:
            self.switch_frame(start)

    def update_display_profile(self) -> None:
        if sys.platform == 'win32':
            if _imagingcms is None:
                return

            assert self.app

            screen_name = self.app.screens()[self.current_screen].name()

            dc = win32gui.CreateDC(screen_name, None, None)

            logging.info(f'Changed screen: {screen_name}')

            icc_path = _imagingcms.get_display_profile_win32(dc, 1)

            if icc_path is not None:
                with open(icc_path, 'rb') as icc:
                    self.display_profile = QColorSpace.fromIccProfile(icc.read())

        if hasattr(self, 'current_output') and self.current_output is not None and self.display_profile is not None:
            self.switch_frame(self.current_output.last_showed_frame)

    def show_message(self, message: str) -> None:
        self.statusbar.showMessage(
            message, round(float(self.settings.statusbar_message_timeout) * 1000)
        )

    def update_statusbar_output_info(self, output: VideoOutput | None = None) -> None:
        output = output or self.current_output
        fmt = output.source.clip.format
        assert fmt

        self.statusbar.total_frames_label.setText(f'{output.total_frames} frames ')
        self.statusbar.duration_label.setText(f'{output.total_time} ')
        self.statusbar.resolution_label.setText(f'{output.width}x{output.height} ')
        self.statusbar.pixel_format_label.setText(f'{fmt.name} ')
        if output.fps_den != 0:
            self.statusbar.fps_label.setText(
                '{}/{} = {:.3f} fps '.format(output.fps_num, output.fps_den, output.fps_num / output.fps_den)
            )
        else:
            self.statusbar.fps_label.setText('{}/{} fps '.format(output.fps_num, output.fps_den))

    def event(self, event: QEvent) -> bool:
        if event.type() == QEvent.LayoutRequest:
            self.timeline.full_repaint()

        return super().event(event)

    # misc methods
    def showEvent(self, event: QShowEvent) -> None:
        super().showEvent(event)
        self.graphics_view.setSizePolicy(self.EVENT_POLICY)

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.settings.autosave_control.value() != Time(seconds=0):
            self.dump_storage_async()

        self.reload_signal.emit()

    def moveEvent(self, _move_event: QMoveEvent) -> None:
        if self.settings.color_management:
            assert self.app
            screen_number = self.app.desktop().screenNumber(self)
            if self.current_screen != screen_number:
                self.current_screen = screen_number
                self.update_display_profile()

    def __getstate__(self) -> Mapping[str, Any]:
        return super().__getstate__() | {
            'window_settings': WindowSettings(
                self.timeline.mode,
                bytes(cast(bytearray, self.saveGeometry())),
                bytes(cast(bytearray, self.saveState()))
            )
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        # toolbars is singleton, so it initialize itself right in its __setstate__()
        self.window_settings = {}

        try:
            try_load(state, 'window_settings', dict, self)
        except BaseException:
            try_load(state, 'timeline_mode', str, self.window_settings)
            try_load(state, 'window_geometry', bytes, self.window_settings)
            try_load(state, 'window_state', bytes, self.window_settings)

        self.timeline.mode = self.window_settings.timeline_mode
        self.restoreGeometry(self.window_settings.window_geometry)
        self.restoreState(self.window_settings.window_state)



from __future__ import annotations

from vapoursynth import EnvironmentData, EnvironmentPolicy, EnvironmentPolicyAPI, register_policy


class VSPreviewEnvironmentPolicy(EnvironmentPolicy):
    _api: EnvironmentPolicyAPI | None = None
    _current: EnvironmentData | None = None

    def __init__(self) -> None:
        pass

    def on_policy_registered(self, special_api: EnvironmentPolicyAPI) -> None:
        self._api = special_api
        self.reload_core()

    def on_policy_cleared(self) -> None:
        self._current = None

    def get_current_environment(self) -> EnvironmentData | None:
        return self._current

    def set_environment(self, environment: EnvironmentData | None) -> None:
        if environment is not self._current:
            raise RuntimeError("The chosen environment is dead.")

    def is_alive(self, environment: EnvironmentData) -> bool:
        return environment is self._current

    def reload_core(self) -> None:
        if self._api is None:
            new_environment = VSPreviewEnvironmentPolicy()
            self._api = new_environment._api
        if self._current:
            del self._current
        try:
            self._api.unregister_policy()
            self._current = self._api.create_environment()
        except BaseException:
            pass


policy = VSPreviewEnvironmentPolicy()
register_policy(policy)


def get_policy() -> VSPreviewEnvironmentPolicy:
    return policy



from .generalmodel import GeneralModel  # noqa: F401
from .picture_types import PictureTypes  # noqa: F401
from .scening import SceningList, SceningLists  # noqa: F401
from .outputs import Outputs, VideoOutputs, AudioOutputs  # noqa: F401



from __future__ import annotations

import logging
import sys
from asyncio import get_event_loop_policy, get_running_loop
from functools import partial, wraps
from platform import python_version
from string import Template
from typing import Any, Callable, Dict, Mapping, Tuple, Type, TypeVar, cast

import vapoursynth as vs
from vsengine.convert import yuv_heuristic
from pkg_resources import get_distribution
from PyQt5.QtCore import QSignalBlocker
from PyQt5.QtWidgets import QApplication

from ..core import Frame, Time, main_window
from ..core.types.enums import ColorRange, Matrix, Primaries, Transfer

T = TypeVar('T')
S = TypeVar('S')
F_SL = TypeVar('F_SL', bound=Callable)


# it is a BuiltinMethodType at the same time
def qt_silent_call(qt_method: F_SL, *args: Any, **kwargs: Any) -> T:
    # https://github.com/python/typing/issues/213
    qobject = qt_method.__self__
    block = QSignalBlocker(qobject)
    ret = qt_method(*args, **kwargs)
    del block
    return cast(T, ret)


class DeltaTemplate(Template):
    delimiter = '%'


def strfdelta(time: Time, output_format: str) -> str:
    seconds = time.value.seconds % (24 * 3600)
    hours = seconds // 3600
    seconds %= 3600
    minutes = seconds // 60
    seconds %= 60
    milliseconds = time.value.microseconds // 1000

    template = DeltaTemplate(output_format)

    return template.substitute(
        D='{:d}'.format(time.value.days),
        H='{:02d}'.format(hours),
        M='{:02d}'.format(minutes),
        S='{:02d}'.format(seconds),
        Z='{:03d}'.format(milliseconds),
        h='{:d}'.format(hours),
        m='{:2d}'.format(minutes),
        s='{:2d}'.format(seconds)
    )


def fire_and_forget(f: F_SL) -> F_SL:
    @wraps(f)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        try:
            loop = get_running_loop()
        except RuntimeError:
            loop = get_event_loop_policy().get_event_loop()
        return loop.run_in_executor(None, partial(f, *args, **kwargs))
    return cast(F_SL, wrapped)


def set_status_label(label: str) -> Callable[[F_SL], F_SL]:
    def _decorator(func: Callable[..., T]) -> Any:
        @wraps(func)
        def _wrapped(*args: Any, **kwargs: Any) -> T:
            main = main_window()

            main.statusbar.label.setText(label)

            ret = func(*args, **kwargs)

            main.statusbar.label.setText('Ready')

            return ret
        return _wrapped

    return cast(Callable[[F_SL], F_SL], _decorator)


def vs_clear_cache() -> None:
    cache_size = vs.core.max_cache_size
    vs.core.max_cache_size = 1
    for output in list(vs.get_outputs().values()):
        if isinstance(output, vs.VideoOutputTuple):
            output.clip.get_frame(int(main_window().current_output.last_showed_frame or Frame(0)))
            break
    vs.core.max_cache_size = cache_size


def check_versions() -> bool:
    if sys.version_info < (3, 10, 0, 'final', 0):
        logging.warning(
            'VSPreview is not tested on Python versions prior to 3.10, but you have {} {}. Use at your own risk.'
            .format(python_version(), sys.version_info.releaselevel)
        )
        return False

    if get_distribution('PyQt5').version < '5.15':
        logging.warning(
            'VSPreview is not tested on PyQt5 versions prior to 5.15, but you have {}. Use at your own risk.'
            .format(get_distribution('PyQt5').version))
        return False

    if vs.core.version_number() < 59:
        logging.warning(
            'VSPreview is not tested on VapourSynth versions prior to 59, but you have {}. Use at your own risk.'
            .format(vs.core.version_number())
        )
        return False

    return True


def get_temp_screen_resolution() -> Tuple[int, int]:
    app = QApplication(sys.argv)

    geometry = app.desktop().screenGeometry()
    return (geometry.width(), geometry.height())


def get_prop(frame: vs.VideoFrame | vs.FrameProps, key: str, t: Type[T]) -> T:
    props = frame if isinstance(frame, Mapping) else frame.props

    try:
        prop = props[key]
    except KeyError:
        raise KeyError(f'get_prop: Key {key} not present in props!')

    if not isinstance(prop, t):
        raise ValueError(f'get_prop: Key {key} did not contain expected type: Expected {t}, got {type(prop)}!')

    return prop


def video_heuristics(
    clip: vs.VideoNode, props: vs.FrameProps | None = None, prop_in: bool = True, string_only: bool = False
) -> Dict[str, str]:
    heuristics = cast(Dict[str, str], yuv_heuristic(clip.width, clip.height))

    if props:
        resize_props = {
            'matrix_in_s': Matrix[get_prop(props, '_Matrix', int)] if '_Matrix' in props else None,
            'primaries_in_s': Primaries[get_prop(props, '_Primaries', int)] if '_Primaries' in props else None,
            'range_in_s': ColorRange[get_prop(props, '_ColorRange', int)] if '_ColorRange' in props else None,
            'transfer_in_s': Transfer[get_prop(props, '_Transfer', int)] if '_Transfer' in props else None,
        }

        if not string_only:
            for key in resize_props:
                if resize_props[key] == 'unspec':
                    resize_props[key] = None

    if not prop_in or string_only:
        heuristics = {f'{k[:-5]}{"" if string_only else "_s"}': v for k, v in heuristics.items()}

    if props:
        heuristics.update({k: v for (k, v) in resize_props.items() if v})

    return heuristics



from __future__ import annotations

from typing import Dict, Protocol, Union, overload

import vapoursynth as vs


class CustomEnum(type):
    items: Dict[str, int]
    values: Dict[int, str]

    @overload
    def __getattribute__(cls, item: str) -> int: ...
    @overload
    def __getattribute__(cls, item: int) -> str: ...

    def __getattribute__(cls, item: str | int) -> int | str:
        return _getattribute(cls, item)

    @overload
    def __getitem__(cls, item: int) -> str: ...
    @overload
    def __getitem__(cls, item: str) -> int: ...

    def __getitem__(cls, item: int | str) -> str | int:
        return _getattribute(cls, item)


def _getattribute(cls: CustomEnum, item: str | int) -> str | int:
    if isinstance(item, str):
        cvalue = object.__getattribute__(cls, item.upper())

        items: Dict[str, int] = object.__getattribute__(cls, 'items')

        if cvalue is not None and cvalue in items:
            return int(items[cvalue])

        item = item.lower()

        if item in items:
            return int(items[item])
    else:
        values: Dict[int, str] = object.__getattribute__(cls, 'values')
        if item in values:
            return str(values[item])

    raise KeyError


_dataT = Union[str, bytes, bytearray, None]


class _ResizerType(Protocol):
    def __call__(
        self, clip: vs.VideoNode, width: int | None, height: int | None, format: int | None, matrix: int | None,
        matrix_s: _dataT, transfer: int | None, transfer_s: _dataT, primaries: int | None, primaries_s: _dataT,
        range: int | None, range_s: _dataT, chromaloc: int | None, chromaloc_s: _dataT, matrix_in: int | None,
        matrix_in_s: _dataT, transfer_in: int | None, transfer_in_s: _dataT, primaries_in: int | None,
        primaries_in_s: _dataT, range_in: int | None, range_in_s: _dataT, chromaloc_in: int | None,
        chromaloc_in_s: _dataT, filter_param_a: float | None, filter_param_b: float | None, resample_filter_uv: _dataT,
        filter_param_a_uv: float | None, filter_param_b_uv: float | None, dither_type: _dataT, cpu_type: _dataT,
        prefer_props: int | None, src_left: float | None, src_top: float | None, src_width: float | None,
        src_height: float | None, nominal_luminance: float | None
    ) -> vs.VideoNode:
        ...


class Matrix(metaclass=CustomEnum):
    values = {
        0: 'rgb',
        1: '709',
        2: 'unspec',
        3: 'reserved',
        4: 'fcc',
        5: '470bg',
        6: '170m',
        7: '240m',
        8: 'ycgco',
        9: '2020ncl',
        10: '2020cl',
        11: 'reserved',
        12: 'chromancl',
        13: 'chromacl',
        14: 'ictcp',
    }
    items = {k: v for v, k in values.items()}

    RGB = values[0]
    BT709 = values[1]
    UNSPEC = values[2]
    BT470_BG = values[5]
    ST170_M = values[6]
    ST240_M = values[7]
    FCC = values[4]
    YCGCO = values[8]
    BT2020_NCL = values[9]
    BT2020_CL = values[10]
    CHROMA_CL = values[13]
    CHROMA_NCL = values[12]
    ICTCP = values[14]


class Transfer(metaclass=CustomEnum):
    values = {
        0: 'reserved',
        1: '709',
        2: 'unspec',
        3: 'reserved',
        4: '470m',
        5: '470bg',
        6: '601',
        7: '240m',
        8: 'linear',
        9: 'log100',
        10: 'log316',
        11: 'xvycc',  # IEC 61966-2-4
        12: 'reserved',
        13: 'srgb',  # IEC 61966-2-1
        14: '2020_10',
        15: '2020_12',
        16: 'st2084',
        17: 'st428',  # not supported by zimg 2.8
        18: 'std-b67',
    }
    items = {k: v for v, k in values.items()}

    BT709 = values[1]
    UNSPEC = values[2]
    BT601 = values[6]
    LINEAR = values[8]
    BT2020_10 = values[14]
    BT2020_12 = values[15]
    ST240_M = values[7]
    BT470_M = values[4]
    BT470_BG = values[5]
    LOG_100 = values[9]
    LOG_316 = values[10]
    ST2084 = values[16]
    ARIB_B67 = values[18]
    SRGB = values[13]
    XV_YCC = values[11]
    IEC_61966_2_4 = XV_YCC
    IEC_61966_2_1 = SRGB


class Primaries(metaclass=CustomEnum):
    values = {
        0: 'reserved',
        1: '709',
        2: 'unspec',
        3: 'reserved',
        4: '470m',
        5: '470bg',
        6: '170m',
        7: '240m',
        8: 'film',
        9: '2020',
        10: 'st428',  # or 'xyz'
        11: 'st431-2',
        12: 'st431-1',
        22: 'jedec-p22',
    }
    items = {k: v for v, k in values.items()}

    BT709 = values[1]
    UNSPEC = values[2]
    ST170_M = values[6]
    ST240_M = values[7]
    BT470_M = values[4]
    BT470_BG = values[5]
    FILM = values[8]
    BT2020 = values[9]
    ST428 = values[10]
    XYZ = ST428
    ST431_2 = values[11]
    ST431_1 = values[12]
    JEDEC_P22 = values[22]
    EBU3213_E = JEDEC_P22


class ColorRange(metaclass=CustomEnum):
    values = {
        0: 'full',
        1: 'limited'
    }
    items = {k: v for v, k in values.items()}

    LIMITED = values[1]
    FULL = values[0]


class ChromaLocation(metaclass=CustomEnum):
    values = {
        0: 'left',
        1: 'center',
        2: 'top_left',
        3: 'top',
        4: 'bottom_left',
        5: 'bottom',
    }
    items = {k: v for v, k in values.items()}

    LEFT = values[0]
    CENTER = values[1]
    TOP_LEFT = values[2]
    TOP = values[3]
    BOTTOM_LEFT = values[4]
    BOTTOM = values[5]



from __future__ import annotations

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLabel

from ...core import AbstractToolbarSettings, Frame, HBoxLayout, SpinBox
from ...main.settings import MainSettings


class PlaybackSettings(AbstractToolbarSettings):
    __slots__ = ('buffer_size_spinbox', )

    CHECKERBOARD_ENABLED = True
    CHECKERBOARD_TILE_COLOR_1 = Qt.white
    CHECKERBOARD_TILE_COLOR_2 = Qt.lightGray
    CHECKERBOARD_TILE_SIZE = 8  # px
    FPS_AVERAGING_WINDOW_SIZE = Frame(100)
    FPS_REFRESH_INTERVAL = 150  # ms
    SEEK_STEP = 1

    def setup_ui(self) -> None:
        super().setup_ui()

        self.buffer_size_spinbox = SpinBox(self, 1, MainSettings.get_usable_cpus_count())

        HBoxLayout(self.vlayout, [QLabel('Playback buffer size (frames)'), self.buffer_size_spinbox])

    def set_defaults(self) -> None:
        self.buffer_size_spinbox.setValue(MainSettings.get_usable_cpus_count())

    @property
    def playback_buffer_size(self) -> int:
        return self.buffer_size_spinbox.value()
